
local o = QuickChatlink;
local m = {};
local m_metatable = { __index = m; };

-- AutoCompletion_t_Init: ()
-- AutoCompletion_t_Hook_ChatEdit_CustomTabPressed: ()
--
-- AutoCompletion_RegisterEditBox: (eb)
-- AutoCompletion_h_EditBox_OnChar: (eb, char)
-- AutoCompletion_h_EditBox_OnHide: (eb)
-- AutoCompletion_h_EditBox_OnTabPressed: (eb)
--
-- m:OnChar: (char)
-- m:OnTabPressed: ()
--
-- m:OnLinkTextChanged: ()
-- m:UpdateSectionIncluded: ()
--
-- m:BeginLink: ()
-- m:FinishLink: ()
-- m:FinishAllPotentialLinks: ()
	-- AutoCompletion_ReplaceTextWithLink: replacement = (leadingChars, linkWithBrackets, followingChars)
-- m:Clear: ()
--
-- m:ScrollToNextSection: ()
-- m:ScrollToNextPotentialMatch: ()

-- config = QuickChatlink_Config;




function o.AutoCompletion_t_Init()
	o.AutoCompletion_t_Init = nil;
	
	o.AutoCompletion_t_Hook_ChatEdit_CustomTabPressed();
end




function o.AutoCompletion_t_Hook_ChatEdit_CustomTabPressed()
	o.AutoCompletion_t_Hook_ChatEdit_CustomTabPressed = nil;
	
	local l_orig_ChatEdit_CustomTabPressed = ChatEdit_CustomTabPressed;
	ChatEdit_CustomTabPressed = function(self)
		-- If the function returns true (meaning an action was taken), normal tab behaviour (cycling through whisper targets) will NOT be performed.
		local didSomething = l_orig_ChatEdit_CustomTabPressed(self);
		if (didSomething ~= true) then
			-- "this" is the edit box in this case.
			local instance = (self or this).QuickChatlink;
			if (instance ~= nil and instance.linkIsOpen == true) then
				instance:OnTabPressed();
				didSomething = true;
			end
		end
		return didSomething;
	end
end




function o.AutoCompletion_RegisterEditBox(eb)
	if (eb.QuickChatlink == nil) then
		local instance = setmetatable({
			parent = eb;
			orig_OnChar = eb:GetScript("OnChar");
			orig_OnHide = eb:GetScript("OnHide");
			includedData = {};
			sectionsNeedingUpdate = {};
		}, m_metatable);
		eb:SetScript("OnChar", o.AutoCompletion_h_EditBox_OnChar);
		eb:SetScript("OnHide", o.AutoCompletion_h_EditBox_OnHide);
		if (eb ~= ChatFrameEditBox) then
			instance.orig_OnTabPressed = eb:GetScript("OnTabPressed");
			eb:SetScript("OnTabPressed", o.AutoCompletion_h_EditBox_OnTabPressed);
		end
		eb.QuickChatlink = instance;
	end
end



function o.AutoCompletion_h_EditBox_OnChar(eb, char)
	local instance = eb.QuickChatlink;
	local orig = instance.orig_OnChar;
	if (orig ~= nil) then orig(eb, char); end
	instance:OnChar(char);
end



function o.AutoCompletion_h_EditBox_OnHide(eb)
	local instance = eb.QuickChatlink;
	instance:Clear();
	local orig = instance.orig_OnHide;
	if (orig ~= nil) then orig(eb); end
end



function o.AutoCompletion_h_EditBox_OnTabPressed(eb)
	local instance = eb.QuickChatlink;
	local orig = instance.orig_OnTabPressed;
	if (orig ~= nil) then orig(eb); end
	instance:OnTabPressed();
end




function m:OnChar(char)
	if (char == "[") then
		self:BeginLink();
	else
		if (self.linkIsOpen == true) then
			-- AceTab-2.0 40685 inserts \255 and \0 to find the location of the cursor and to erase that beacon when the cursor is found, respectively.
			if (char ~= "\255" and char ~= "\0") then
				self:OnLinkTextChanged();
			end
		else
			if (char == "]") then
				self:FinishAllPotentialLinks();
			end
		end
	end
end



function m:OnTabPressed()
	if (self.linkIsOpen ~= true) then
		return;
	end
	if (IsControlKeyDown() ~= nil) then
		self:ScrollToNextSection();
	elseif (IsShiftKeyDown() ~= nil) then
		self:ScrollToNextPotentialMatch();
	else
		self:FinishLink();
	end
end




function m:OnLinkTextChanged()
	local text = self.parent:GetText();
	local linkBeginPos = text:reverse():find("[", 1, true);
	-- Don't go forward if the bracket is missing. This can happen if the user deletes the left bracket.
	if (linkBeginPos == nil) then
		self:Clear();
		return;
	end
	local isComplete = (text:sub(-1, -1) == "]");
	-- Add 2 because 1) we found the position of the opening bracket, not the position of the letter after it;
	-- and 2) strings are 1-based, but text:len() - linkBeginPos gives an offset relative to 0.
	local newLinkText = text:sub(((text:len() - linkBeginPos) + 2), ((isComplete == true and -2) or -1));
	
	local oldPartialLinkText = self.partialLinkText;
	-- Don't go forward if the text in front of the bracket has not changed. This can happen if the user goes back and
	-- types text into the message before the start of the current link. The exception to this is when the link has been
	-- completed by the user manually entering the closing right bracket, in which case we still need to do an update so
	-- that we've narrowed the included data using the end-of-string anchor ($).
	if (oldPartialLinkText == newLinkText and isComplete == false) then
		return;
	end
	self.partialLinkText = newLinkText;
	self.isComplete = isComplete;
	
	-- If this is true, then the user had entered "item" but then set the text to "iten" or "otem" or this is the first character in the link.
	-- Regardless of the cause, if it's true, we must look through the base DBs to add any matching links to the included list.
	local needsExpanding = (oldPartialLinkText ~= newLinkText:sub(1, oldPartialLinkText:len()) or oldPartialLinkText == "");
	local sectionsNeedingUpdate = self.sectionsNeedingUpdate;
	
	for index, included in pairs(self.includedData) do
		if (included ~= false) then
			-- Set this flag. If it's true, it means it needs expanding. If it's false, it means it doesn't need expanding but does need narrowing.
			-- Only set it if it's not already true, because if the new flag is false, it could overwrite the fact that this particular section requires expanding.
			if (sectionsNeedingUpdate[index] ~= true) then
				sectionsNeedingUpdate[index] = needsExpanding;
			end
		end
	end
	
	local currSectionIndex = self.currSectionIndex;
	if (currSectionIndex == nil) then
		-- If currSectionIndex is nil, it means we previously determined that there were no matches available in any section.
		-- But if we just set the needsExpanding, maybe there will be matches once the updates are performed. The scroll
		-- function performs these updates when it's looking for a new section with matching links, so we'll call it.
		if (needsExpanding == true) then
			self:ScrollToNextSection();
		end
	else
		-- This code doesn't make sense to include, because for some letters, just the first one IS enough narrowing.
		-- For instance, "a" is not enough narrowing, but "x" certainly is.
		-- The proper approach is the new double-bracket opener ([[), which lets the user decide when they want autocomplete active.
		--[[
		-- Only allow the flag to be set for the current section if the number of letters typed is 2 or more.
		-- Updating before then doesn't provide enough meaningful narrowing.
		if (newLinkText:len() > 1) then
			self:UpdateSectionIncluded(currSectionIndex);
		end
		--]]
		self:UpdateSectionIncluded(currSectionIndex);
		local currSectionIncluded = self.includedData[currSectionIndex];
		if (next(currSectionIncluded) == nil) then
			-- If there's nothing in the current section, we should try the next section. Even if we didn't just expand, we should
			-- still check, since self.currSectionIndex will go nil once we've determined that there's nothing in any section.
			self:ScrollToNextSection();
		else
			-- There is stuff in the current section, but maybe we just removed the currently selected key.
			-- If so, we should try to find the first potential match in this section and scroll to that.
			if (self.currKey == nil or currSectionIncluded[self.currKey] == nil) then
				self.currKey = nil;
				self.currValue = nil;
				self:ScrollToNextPotentialMatch();
			else
				self:UpdateHighlightedText();
			end
		end
	end
	
	if (isComplete == true) then
		self:FinishLink();
	end
end



function m:UpdateSectionIncluded(index)
	local matchingText = ("^" .. self.partialLinkText:NicheDevTools_EscapePatterns():NicheDevTools_CaseInsensitize());
	if (self.isComplete == true) then
		matchingText = (matchingText .. "$");
	end
	local funcs = o.Sections_sections[o.config.order[index]];
	local matcher, matchFunc, mismatchVal = funcs.CreateMatcher(matchingText);
	local included = self.includedData[index];
	
	for linkID, data in pairs(included) do
		if (matchFunc(data, matcher) == mismatchVal) then
			included[linkID] = nil;
		end
	end
	
	-- true is just the flag that says it needs expansion.
	if (self.sectionsNeedingUpdate[index] == true) then
		for linkID, data in funcs.IterateDB(false, false) do
			if (matchFunc(data, matcher) ~= mismatchVal) then
				included[linkID] = data;
			end
		end
	end
	
	self.sectionsNeedingUpdate[index] = nil;
end




function m:BeginLink()
	if (o.config.noDoubleBracket == nil) then
		if (self.parent:GetText():sub(-2, -1) ~= "[[") then
			return;
		end
	else
		if (self.parent:GetText():sub(-1, -1) ~= "[") then
			return;
		end
	end
	if (self.linkIsOpen == true) then
		self:Clear();
	end
	
	local included = self.includedData;
	local sectionsNeedingUpdate = self.sectionsNeedingUpdate;
	local sections = o.Sections_sections;
	for index, name in pairs(o.config.order) do
		if (sections[name] ~= nil) then
			included[index] = {};
			if (self.currSectionIndex == nil) then
				self.currSectionIndex = index;
			end
		else
			included[index] = false;
		end
		sectionsNeedingUpdate[index] = nil;
	end
	self.partialLinkText = ("");
	self.isComplete = nil;
	self.currKey = nil;
	self.currValue = nil;
	self.linkIsOpen = true;
end



function m:FinishLink()
	if (self.linkIsOpen == true and self.currKey ~= nil) then
		local link = o.Sections_sections[o.config.order[self.currSectionIndex]].BuildChatlink(self.currKey, self.currValue);
		if (link ~= nil) then
			local eb = self.parent;
			local text = eb:GetText();
			local linkBeginPos = text:reverse():find("[", 1, true);
			if (o.config.noDoubleBracket == nil) then
				-- Add 1 to account for the double bracket.
				linkBeginPos = (linkBeginPos + 1);
			end
			eb:SetText(text:sub(1, (text:len() - linkBeginPos)) .. link);
		end
	end
	
	self:Clear();
end



function m:FinishAllPotentialLinks()
	if (o.config.noNormalReplacement == nil) then
		local openingExtra;
		if (o.config.noDoubleBracket == nil) then
			openingExtra = ("%[?");
		end
		local closingExtra;
		if (o.config.normalClosingDoubleBracket == true) then
			closingExtra = ("%]");
		end
		local pattern = ("(|?h?%s)(%%b[])(|?h?%s)"):format((openingExtra or ""), (closingExtra or ""));
		local eb = self.parent;
		eb:SetText((eb:GetText():gsub(pattern, o.AutoCompletion_ReplaceTextWithLink)));
	end
end


function o.AutoCompletion_ReplaceTextWithLink(leadingChars, linkWithBrackets, followingChars)
	-- Make sure it's not already a link.
	if (leadingChars == "|h" or followingChars == "|h") then
		return nil;
	end
	
	local link = nil;
	local matchingText = ("^" .. linkWithBrackets:sub(2, -2):NicheDevTools_EscapePatterns():NicheDevTools_CaseInsensitize() .. "$");
	local sections = o.Sections_sections;
	local order = o.config.order;
	
	local index = 1;
	local name = order[1];
	local funcs, matcher, matchFunc, mismatchVal;
	local iterator, state, start;
	local linkID, data;
	while (name ~= nil and link == nil) do
		funcs = sections[name];
		if (funcs ~= nil) then
			matcher, matchFunc, mismatchVal = funcs.CreateMatcher(matchingText);
			iterator, state, start = funcs.IterateDB(false, false);
			linkID, data = iterator(state, start);
			while (linkID ~= nil and link == nil) do
				if (matchFunc(data, matcher) ~= mismatchVal) then
					link = funcs.BuildChatlink(linkID, data);
				end
				linkID, data = iterator(state, linkID);
			end
		end
		if (link == nil) then
			index = (index + 1);
			name = order[index];
		end
	end
	
	if (link ~= nil) then
		-- Account for places where the user typed the double-bracket opening syntax ([[link]) instead of normal single-bracket.
		if (o.config.noDoubleBracket == nil and leadingChars:sub(-1, -1) == "[") then
			leadingChars = leadingChars:sub(1, -2);
		end
		-- Account for places where the user typed the double-bracket closing syntax ([link]]) instead of normal single-bracket.
		if (o.config.normalClosingDoubleBracket == true and followingChars:sub(1, 1) == "]") then
			followingChars = followingChars:sub(2, -1);
		end
		link = (leadingChars .. link .. followingChars);
	end
	return link;
end



function m:Clear()
	if (self.linkIsOpen ~= true) then
		return;
	end
	
	local included = self.includedData;
	local sectionsNeedingUpdate = self.sectionsNeedingUpdate;
	for index = 1, #included, 1 do
		included[index] = nil;
		sectionsNeedingUpdate[index] = nil;
	end
	self.partialLinkText = nil;
	self.isComplete = nil;
	self.currSectionIndex = nil;
	self.currKey = nil;
	self.currValue = nil;
	self.linkIsOpen = nil;
end




function m:ScrollToNextSection()
	local found = false;
	local next = next;
	local includedData = self.includedData;
	local sectionsNeedingUpdate = self.sectionsNeedingUpdate;
	local currSectionIndex = self.currSectionIndex;
	
	-- First look from one past our current index up to the end of the included array.
	local maxIndex = #includedData;
	local index = ((currSectionIndex or 0) + 1);
	local sectionIncluded;
	while (found == false and index <= maxIndex) do
		if (sectionsNeedingUpdate[index] ~= nil) then
			-- We'd better update this section now so we know whether it has any data.
			self:UpdateSectionIncluded(index);
		end
		sectionIncluded = includedData[index];
		-- Consider this section invalid if it has nothing in it.
		if (sectionIncluded == false or next(sectionIncluded) == nil) then
			index = (index + 1);
		else
			found = true;
		end
	end
	
	-- Next, if we found nothing before, try searching from the first index up to our starting index.
	-- Intentionally, scrolling to the "next" section may end up scrolling to the same section we
	-- started at, if there is nothing in any other section. This is okay. If there was nothing in the
	-- starting section, we will just end up with self.currSectionIndex being nil, like we started with.
	if (found == false and currSectionIndex ~= nil) then
		maxIndex = currSectionIndex;
		index = 1;
		while (found == false and index <= maxIndex) do
			if (sectionsNeedingUpdate[index] ~= nil) then
				-- We'd better update this section now so we know whether it has any data.
				self:UpdateSectionIncluded(index);
			end
			sectionIncluded = includedData[index];
			-- Consider this section invalid if it has nothing in it.
			if (sectionIncluded == false or next(sectionIncluded) == nil) then
				index = (index + 1);
			else
				found = true;
			end
		end
	end
	
	self.currKey, self.currValue = nil, nil;
	if (found == false) then
		-- We found nothing valid at all in any section.
		self.currSectionIndex = nil;
	else
		self.currSectionIndex = index;
		self:ScrollToNextPotentialMatch();
	end
end



function m:ScrollToNextPotentialMatch()
	if (self.currSectionIndex == nil) then
		-- This can only happen when there is nothing valid at all in any section.
		return;
	end
	
	local nextKey, nextValue = next(self.includedData[self.currSectionIndex], self.currKey);
	if (nextKey ~= nil) then
		self.currKey, self.currValue = nextKey, nextValue;
		self:UpdateHighlightedText();
	else
		-- Ran out of keys in this section. Try the next section.
		self:ScrollToNextSection();
	end
end




function m:UpdateHighlightedText()
	local linkName = o.Sections_sections[o.config.order[self.currSectionIndex]].GetPlainName(self.currKey, self.currValue);
	local eb = self.parent;
	local text = self.parent:GetText();
	local partialLen = self.partialLinkText:len();
	-- (text:len() - linkBeginPos) puts us just before the opening bracket. We want to include the bracket, so we add 1.
	-- Adding partialLen puts us on the first character before the untyped portion of the link name.
	local endPos = ((text:len() - text:reverse():find("[", 1, true)) + 1 + partialLen);
	eb:SetText(text:sub(1, endPos) .. linkName:sub((partialLen + 1), -1) .. "]");
	-- The highlight is 0-based, so don't add 1 here.
	eb:HighlightText(endPos);
end
